;;; input.el --- The heart of external version of Aldor Scanner

;; Author: Jeremy (Ya-Yu) Hsieh <hsiehy@unbc.ca>
;; Maintainer: None

;; This file contains the code for external version Scanner
;; for Aldor


;;; Commentary:

;; This file contains the code and implementation for the
;; external version Scanner for Aldor.  While I am working
;; on this file, one of the main challenge is the timing
;; problem.  Every micro-second for computer may make a difference
;; for the result and output.  Therefore, I spend most of my
;; time on planning the program, and debugging the timing
;; issue.  As the functions and variables used in this file,
;; I will do the comment alone with the code.


;;; Note:

;; For more detail of the logic and information, please refer
;; to my documentation/report/thesis





;; "scanner-done" variable is declare as global variable
;; the propose of this variable is to determine if the
;; scanner had finish scanning the whole file AND if all
;; of the information in middle.txt had been process.

(defvar aldor-scanner-done nil)





;; "aldor-process-done-sentinel" uses as the sentinel to determine
;; if the state of process had been changed.  If the state
;; had been changed, and had been changed as any of the following
;; state - exit, signal, close, or nil; then set the variable
;; "aldor-scanner-done" to t (true).  For more detail about sentinel
;; please refer to my documentation/report/Thesis or look up in
;; Emacs lisp manual (C-h i -> Emacs -> Emacs Lisp -> Sentinels) 

(defun aldor-process-done-sentinel (process string)
  (when (memq (process-status process) '(exit signal closed nil))
    (setq aldor-scanner-done t)))





;; "aldor-restore-point-filter" is the Filter used for restore the
;; point to the original place.  "save-excursion" will remember
;; the original state.  The logic is go to the wherever the
;; process-mark of process is currently at. (Normally is at the
;; end of the buffer). The function then inserts the string, the 
;; side effect is to move the point.  At last, set the process-marker
;; of the process to wherever the point is.  After it exit the
;; function, the "save-excursion" will restore the point to wherever
;; position before enter the filter.  Therefore, the point will NOT
;; get mess up.  (For more information about the Filter, please
;; refer to my document/report OR look up into the Emacs Manual
;; C-h i -> Emacs -> Emacs Lisp -> Filters) 

(defun aldor-restore-point-filter (proc string)
  (with-current-buffer (process-buffer proc)
    (save-excursion
      (goto-char (process-mark proc))
      (insert string)
      (set-marker (process-mark proc) (point)))))





;; "aldor-read-from-list" takes two argument, proc and result
;; "proc" is the process of scanner (scanner.exe).
;; "result" is the buffer where the final result should write to.
;; (Where the source file is).
;; This function uses the function aldor-get-1-value to read the information
;; from the intermediate buffer (middle.txt).  Afterward, it calls the
;; function aldor-colour-syntax to do the colour syntax to the source/result
;; buffer. 
;;
;; Note: result file and source file is the SAME.  Although they are 
;;       the same buffer, but I use two different variable to reduce the
;;       confusion.  Source means original buffer; result means modified
;;       buffer.

(defun aldor-read-from-list (proc result)
  (let ((marker (make-marker))
	(buffer (process-buffer proc)))	
    (with-current-buffer buffer
      (beginning-of-buffer)
      (set-marker marker (point) buffer)
      (set-marker-insertion-type marker nil))
    (let ((l (aldor-get-1-value buffer)))
      (while (not (or (equal l 'Process)
		      (and aldor-scanner-done (null l))))
	(aldor-colour-syntax (nth 0 l)
			     (nth 1 l)
			     (nth 2 l)
			     result)
	(setq l (aldor-get-1-value marker))))))





;; "aldor-colour-syntax" function does the face colour change for the result
;; buffer.  This function takes four arguments -
;;    - start: The start point of the token
;;    - end  : The end point of the token
;;    - type : Type of the token, so it can do the different colour change
;;    - result: The buffer where the colour should be apply to
;;
;; Note:
;;    - This function change the buffer colour ONLY.  Therefore, this 
;;      function does NOT save the file.  Therefore, if the user
;;      does not save the buffer, then the change will be erase.
;;      (Primary goal is to avoid change the user's file)
;;    - Set def-op-word to the same face - keyword face
;;    - Set aldor-re-doc and aldor-post-doc has same color - doc face
;;    - Set float and int has same color - constant face
;;    - Set operators (def, res, & future)to  same color - buildin face

(defun aldor-colour-syntax( start end type result)
  (cond ((equal type 'aldor-reserved)
	 (put-text-property start end 'face 'font-lock-keyword-face result)
	 (put-text-property start end 'help-echo "RESERVED" result))
	((equal type 'aldor-def-op-word)
	 (put-text-property start end 'face 'font-lock-keyword-face result)
	 (put-text-property start end 'help-echo "DEFINABLE KEYWORD" result))
	((equal type 'aldor-class)
	 (put-text-property start end 'face 'font-lock-type-face result)
	 (put-text-property start end 'help-echo "CLASS" result))
	((equal type 'aldor-import-function)
	 (put-text-property start end 'face 'font-lock-function-name-face result)
	 (put-text-property start end 'help-echo "IMPORT FUNCTION" result))
	((equal type 'aldor-pre-doc)
	 (put-text-property start end 'face 'font-lock-doc-face result)
	 (put-text-property start end 'help-echo "PRE-DOCUMENT" result))
	((equal type 'aldor-post-doc)
	 (put-text-property start end 'face 'font-lock-doc-face result)
	 (put-text-property start end 'help-echo "POST-DOCUMENT" result))
	((equal type 'aldor-comment)
	 (put-text-property start end 'face 'font-lock-comment-face result)
	 (put-text-property start end 'help-echo "COMMENT" result))
	((equal type 'aldor-string)
	 (put-text-property start end 'face 'font-lock-string-face result)
	 (put-text-property start end 'help-echo "STRING" result))
	((equal type 'aldor-id)
	 (put-text-property start end 'face 'font-lock-variable-name-face result)
	 (put-text-property start end 'help-echo "IDENTIFIER" result))
	((equal type 'aldor-float)
	 (put-text-property start end 'face 'font-lock-constant-face result)
	 (put-text-property start end 'help-echo "FLOAT" result))
	((equal type 'aldor-int)
	 (put-text-property start end 'face 'font-lock-constant-face result)
	 (put-text-property start end 'help-echo "INTEGER" result))
	((equal type 'aldor-def-op)
	 (put-text-property start end 'face 'font-lock-builtin-face result)
	 (put-text-property start end 'help-echo "DEFINABLE OPERATOR" result))
	((equal type 'aldor-res-op)
	 (put-text-property start end 'face 'font-lock-builtin-face result)
	 (put-text-property start end 'help-echo "RESERVED OPERATOR" result))
	((equal type 'aldor-future-op)
	 (put-text-property start end 'face 'font-lock-buildin-face result)
	 (put-text-property start end 'help-echo "FUTURE OPERATOR" result))))





;; The main program to call, which should do ALL the works
;;   - add the path
;;   - set all the variables (source buffer, result buffer, and process)
;;   - erase the source buffer (in case of unwanted information in buffer)
;;   - called "scanner" program outside of emacs
;;   - called "set-process-sentinel" to finalize the done flag
;;   - called "aldor-restore-point-filter" to set the point to proper place
;;   - called "aldor-read-from-list" to read the information from source buffer
;;     (Pass in the process and result buffer
;;

(defun aldor-scan (aldor-file-name)
  (interactive "fFile to scan? ")
  (let
      ((exec-path (cons default-directory exec-path))
       (buffer (get-buffer-create "middle.txt"))
       (result (find-file-noselect aldor-file-name))
       (proc1 nil))
    (with-current-buffer buffer
      (erase-buffer))
    (setq aldor-scanner-done nil)
    (setq proc1 (start-process "token" buffer "aldor-scanner" aldor-file-name))
    (set-process-sentinel proc1 'aldor-process-done-sentinel)
    (set-process-filter proc1 'aldor-restore-point-filter)
    (aldor-read-from-list proc1 result)))





;; "aldor-get-1-value" read from the buffer (middle.txt) and return
;; one value at once.  For more detail, please refer to my
;; report/document

(defun aldor-get-1-value (buffer)
  (let ((value
	 (condition-case nil
	     (read buffer)
	   (error nil)))
	(done nil))
    (while (and (not value) (not done))
      (setq done (or aldor-scanner-done
		     (null (accept-process-output nil 1))
		     (sit-for 100.0)
		     ))
      (setq value
	    (condition-case nil
		(read buffer)
	      (error nil))))
    value))

(provide 'aldor-external)